// This file is part of CAF, the C++ Actor Framework. See the file LICENSE in
// the main distribution directory for license terms and copyright or visit
// https://github.com/actor-framework/actor-framework/blob/master/LICENSE.

#pragma once

#include "caf/ipv6_endpoint.hpp"

namespace caf {

/// An IP endpoint that contains an ::ipv6_address and a port.
using ip_endpoint = ipv6_endpoint;

} // namespace caf
