// This file is part of CAF, the C++ Actor Framework. See the file LICENSE in
// the main distribution directory for license terms and copyright or visit
// https://github.com/actor-framework/actor-framework/blob/master/LICENSE.

#pragma once

namespace caf::io::basp {

struct header;

class worker;
class worker_hub;
class message_queue;
class instance;
class routing_table;

} // namespace caf::io::basp
