NCP_FUNCTION(NCP_FILE_SET_LOCK,			0x01)
NCP_FUNCTION(NCP_FILE_RELEASE_LOCK,		0x02)
NCP_FUNCTION(NCP_LOG_FILE,			0x03)
NCP_FUNCTION(NCP_LOCK_FILE_SET,			0x04)
NCP_FUNCTION(NCP_RELEASE_FILE,			0x05)
NCP_FUNCTION(NCP_RELEASE_FILE_SET,		0x06)
NCP_FUNCTION(NCP_CLEAR_FILE,			0x07)
NCP_FUNCTION(NCP_CLEAR_FILE_SET,		0x08)
NCP_FUNCTION(NCP_LOG_LOGICAL_RECORD,		0x09)
NCP_FUNCTION(NCP_LOCK_LOGICAL_RECORD_SET,	0x0a)
NCP_FUNCTION(NCP_CLEAR_LOGICAL_RECORD,		0x0b)
NCP_FUNCTION(NCP_RELEASE_LOGICAL_RECORD,	0x0c)
NCP_FUNCTION(NCP_RELEASE_LOGICAL_RECORD_SET,	0x0d)
NCP_FUNCTION(NCP_CLEAR_LOGICAL_RECORD_SET,	0x0e)
NCP_FUNCTION(NCP_ALLOC_RESOURCE,		0x0f)
NCP_FUNCTION(NCP_DEALLOC_RESOURCE,		0x10)
NCP_FUNCTION(NCP_PRINT,				0x11)
NCP_FUNCTION(NCP_MESSAGE,			0x15)
NCP_FUNCTION(NCP_DIRECTORY,			0x16)
NCP_FUNCTION(NCP_BINDARY_AND_MISC,		0x17)
NCP_FUNCTION(NCP_END_OF_JOB,			0x18)
NCP_FUNCTION(NCP_LOGOUT,			0x19)
NCP_FUNCTION(NCP_LOG_PHYSICAL_RECORD,		0x1a)
NCP_FUNCTION(NCP_LOCK_PHYSICAL_RECORD_SET,	0x1b)
NCP_FUNCTION(NCP_RELEASE_PHYSICAL_RECORD,	0x1c)
NCP_FUNCTION(NCP_RELEASE_PHYSICAL_RECORD_SET,	0x1d)
NCP_FUNCTION(NCP_CLEAR_PHYSICAL_RECORD,		0x1e)
NCP_FUNCTION(NCP_CLEAR_PHYSICAL_RECORD_SET,	0x1f)
NCP_FUNCTION(NCP_SEMAPHORE,			0x20)
NCP_FUNCTION(NCP_TRANSACTION_TRACKING, 		0x22)
NCP_FUNCTION(NCP_AFP, 				0x23)
NCP_FUNCTION(NCP_CLOSE_FILE,			0x42)
NCP_FUNCTION(NCP_GET_FILE_SIZE,			0x47)
NCP_FUNCTION(NCP_READ_FILE,			0x48)
NCP_FUNCTION(NCP_WRITE_FILE,			0x49)
NCP_FUNCTION(NCP_EXT_ATTR,			0x56)
NCP_FUNCTION(NCP_FILE_DIR,			0x57)
NCP_FUNCTION(NCP_CREATE_FILE_DIR,		0x5701)
NCP_FUNCTION(NCP_INIT_SEARCH,			0x5702)
NCP_FUNCTION(NCP_SEARCH_FILE_DIR,		0x5703)
NCP_FUNCTION(NCP_RENAME_FILE_DIR,		0x5704)
NCP_FUNCTION(NCP_OBTAIN_FILE_DIR_INFO,		0x5706)
NCP_FUNCTION(NCP_MODIFY_FILE_DIR_DOS_INFO,	0x5707)
NCP_FUNCTION(NCP_DELETE_FILE_DIR,		0x5708)
NCP_FUNCTION(NCP_SET_SHORT_DIR_HANDLE,		0x5709)
NCP_FUNCTION(NCP_SEARCH_FOR_FILE_DIR_SET,	0x5714)
NCP_FUNCTION(NCP_GET_NAME_SPACE_LOADED_LIST,	0x5718)
NCP_FUNCTION(NCP_GET_CURRENT_SIZE_OF_FILE,	0x5742)
NCP_FUNCTION(NCP_AUDITING,			0x58)
NCP_FUNCTION(NCP_MIGRATION,			0x5a)
NCP_FUNCTION(NCP_PNW,				0x60)
NCP_FUNCTION(NCP_GET_MAX_PACKET_SIZE,		0x61)
NCP_FUNCTION(NCP_NDS,				0x68)
NCP_FUNCTION(NCP_SEMAPHORE_NEW,			0x6f)
NCP_FUNCTION(NCP_7B,				0x7b)
