// This file is part of CAF, the C++ Actor Framework. See the file LICENSE in
// the main distribution directory for license terms and copyright or visit
// https://github.com/actor-framework/actor-framework/blob/master/LICENSE.

// This header intentionally has no `#pragma once`. See fsm.hpp.

#undef CAF_FSM_EVAL_ACTION

#undef CAF_FSM_EVAL_MISMATCH_EC

#undef start

#undef state

#undef term_state

#undef unstable_state

#undef fin

#undef CAF_TRANSITION_IMPL1

#undef CAF_TRANSITION_IMPL2

#undef CAF_TRANSITION_IMPL3

#undef CAF_TRANSITION_IMPL4

#undef CAF_ERROR_TRANSITION_IMPL2

#undef CAF_ERROR_TRANSITION_IMPL1

#undef CAF_EPSILON_IMPL1

#undef CAF_EPSILON_IMPL2

#undef CAF_EPSILON_IMPL3

#undef CAF_EPSILON_IMPL4

#undef CAF_FSM_TRANSITION_IMPL2

#undef CAF_FSM_TRANSITION_IMPL3

#undef CAF_FSM_TRANSITION_IMPL4

#undef CAF_FSM_TRANSITION_IMPL5

#undef CAF_FSM_EPSILON_IMPL2

#undef CAF_FSM_EPSILON_IMPL3

#undef CAF_FSM_EPSILON_IMPL4

#undef CAF_FSM_EPSILON_IMPL5

#undef transition

#undef error_transition

#undef epsilon

#undef fsm_transition

#undef fsm_epsilon

#undef transition_if

#undef epsilon_if

#undef fsm_transition_if

#undef fsm_epsilon_if
