// This file is part of CAF, the C++ Actor Framework. See the file LICENSE in
// the main distribution directory for license terms and copyright or visit
// https://github.com/actor-framework/actor-framework/blob/master/LICENSE.

#define CAF_SUITE detail.type_id_list_builder

#include "caf/detail/type_id_list_builder.hpp"

#include "caf/test/dsl.hpp"

using namespace caf;

namespace {

struct fixture {};

} // namespace

CAF_TEST_FIXTURE_SCOPE(type_id_list_builder_tests, fixture)

CAF_TEST(todo) {
  // implement me
}

CAF_TEST_FIXTURE_SCOPE_END()
