#
# Settings that specify the version of ZeekControl
#

VERSION = "@VERSION@"
ZEEKBASE = "@PREFIX@"
CFGFILE = "@ETC@/zeekctl.cfg"
ZEEKSCRIPTDIR = "@ZEEKSCRIPTDIR@"
LIBDIR = "@LIBDIR@"
LIBDIRINTERNAL = "@PY_MOD_INSTALL_DIR@"
