// This file is part of CAF, the C++ Actor Framework. See the file LICENSE in
// the main distribution directory for license terms and copyright or visit
// https://github.com/actor-framework/actor-framework/blob/master/LICENSE.

#pragma once

#include <cstdint>

namespace caf::io::basp {

/// @addtogroup BASP
/// @{

/// The current BASP version. Note: BASP is not backwards compatible.
constexpr uint64_t version = 4;

/// @}

} // namespace caf::io::basp
