// This file is part of CAF, the C++ Actor Framework. See the file LICENSE in
// the main distribution directory for license terms and copyright or visit
// https://github.com/actor-framework/actor-framework/blob/master/LICENSE.

#pragma once

#include "caf/intrusive/inbox_result.hpp"

namespace caf::detail {

/// Alias for backwards compatibility.
using enqueue_result = intrusive::inbox_result;

} // namespace caf::detail
