// This file is part of CAF, the C++ Actor Framework. See the file LICENSE in
// the main distribution directory for license terms and copyright or visit
// https://github.com/actor-framework/actor-framework/blob/master/LICENSE.

#pragma once

#include <string>
#include <utility>
#include <vector>

#include "caf/detail/core_export.hpp"

namespace caf::detail {

CAF_CORE_EXPORT unsigned get_process_id();

} // namespace caf::detail
