// This file is part of CAF, the C++ Actor Framework. See the file LICENSE in
// the main distribution directory for license terms and copyright or visit
// https://github.com/actor-framework/actor-framework/blob/master/LICENSE.

#pragma once

/// Integer value for the QUIET log level.
#define CAF_LOG_LEVEL_QUIET 0

/// Integer value for the ERROR log level.
#define CAF_LOG_LEVEL_ERROR 3

/// Integer value for the WARNING log level.
#define CAF_LOG_LEVEL_WARNING 6

/// Integer value for the INFO log level.
#define CAF_LOG_LEVEL_INFO 9

/// Integer value for the DEBUG log level.
#define CAF_LOG_LEVEL_DEBUG 12

/// Integer value for the TRACE log level.
#define CAF_LOG_LEVEL_TRACE 15
