// This file is part of CAF, the C++ Actor Framework. See the file LICENSE in
// the main distribution directory for license terms and copyright or visit
// https://github.com/actor-framework/actor-framework/blob/master/LICENSE.

#pragma once

namespace caf::tag {

/// Allows the testing DSL to recognize that subtypes are boxing content types.
struct boxing_type {};

} // namespace caf::tag
